#if defined(unix) || defined(__SASC)

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#if defined(_AIX)
#include <sys/select.h>
#endif
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>
#include <fcntl.h>

#elif defined(_WIN32)
#include <winsock2.h>

#endif

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "mud.h"
#include "server.h"

#if defined(_WIN32)
#define PERROR	doPerror
#else
#define PERROR	perror
#endif

#define MACHINE_DEBUG	0

#define MAX_DELAY	1000

typedef enum {
    cs_name,		/* want a player name */
    cs_password1,	/* want a player password */
    cs_password2,	/* first retry */
    cs_password3,	/* second and last retry */
    cs_connected,	/* fully connected */
    cs_createQuery,	/* want to create new one? */
    cs_createPW1,	/* first copy of new player password */
    cs_createPW2	/* second copy of new player password */
} ClientState_t;

typedef struct client {
    struct client *cl_next;
    struct client *cl_prev;
    int cl_fd;
    char *cl_hostName;
    Player_t *cl_player;
    int cl_inputLength;
    ClientState_t cl_state;
    Bool_t cl_exiting;
    Bool_t cl_isMachine;
    Bool_t cl_deleted;
    char cl_inputLine[LINE_LENGTH];
} Client_t;

typedef struct action {
    struct action *ac_next;
    Client_t *ac_client;
    int ac_delay;
    char *ac_action;
} Action_t;

static char *DbPath;
static int Socket;
static Bool_t WantQuit;
static Client_t *Clients;
static fd_set AllPlayerBits;
static int PlayerCount;
static Action_t *Actions;
static int CurrentDelay, LastTime;
static char UserMessage[LINE_LENGTH], PrintMessage[LINE_LENGTH];
static int ScanPos;

static void
killSocket(int s)
{

    shutdown(s, 2);
#if defined(_WIN32)
    closesocket(s);
#elif defined(__SASC)
    CloseSocket(s);
#else
    close(s);
#endif
}

void
mudAbort(char *message)
{

    if (message != NULL) {
	fprintf(stderr, "ToyMud server: *** %s - aborting\n", message);
    }
    killSocket(Socket);
#if defined(_WIN32)
    WSACleanup();
#endif
    exit(1);
}

#if defined(_WIN32)

/* All 'perror's in this file are from socket calls. */

#define ERROR_BUF_LEN	1000

static void
doPerror(const char *message)
{
    HMODULE wsMod;
    DWORD mLen, len, errorCode;
    char buffer[ERROR_BUFF_LEN];

    errorCode = WSAGetLastError();
    sprintf(&buffer[0], "%s: ", message);
    mLen = strlen(&buffer[0]);
    wsMod = GetModuleHandle("wsock32");
    len =
	FormatMessage(FORMAT_MESSAGE_FROM_HMODULE | FORMAT_MESSAGE_FROM_SYSTEM,
		      (LPCVOID) wsMod, errorCode, 0, &buffer[mLen],
		      ERROR_BUFF_LEN - mLen - 1, NULL);
    if (len == 0) {
	sprintf(&buffer[mLen], "FormatMessage failed reason %d, orig error %d",
		GetLastError(), errorCode);
    } else {
	/* trim any trailing CR/LF */
	mLen = strlen(&buffer[0]);
	while (mLen != 0 &&
	       (buffer[mLen - 1] == '\n' || buffer[mLen - 1] == '\r'))
	{
	    mLen -= 1;
	}
	buffer[mLen] = '\0';
    }
    fprintf(stderr, "%s\n", &buffer[0]);
}

#endif

void *
mudAlloc(unsigned int len)
{
    void *p;

    p = malloc(len);
    if (p == NULL) {
	mudAbort("out of memory");
    }
/*printf("mudAlloc(%d) => 0x%08x\n", len, p);fflush(stdout);*/
    return(p);
}

void
mudFree(void *p)
{

/*printf("mudFree(0x%x)\n", p);fflush(stdout);*/
    free(p);
}

char *
strdup(const char *s)
{
    char *p;

    p = mudAlloc(strlen(s) + sizeof(char));
    strcpy(p, s);
    return(p);
}

void
aprintf(Client_t *active, char *fmt, ...)
{
    va_list ap;
    int len;
    Client_t *cl;
    Player_t *pl;
    Thing_t *here;

    va_start(ap, fmt);
    vsprintf(&PrintMessage[0], fmt, ap);
    va_end(ap);
    len = strlen(&PrintMessage[0]);
    if (len >= LINE_LENGTH) {
	mudAbort("message too long in aprintf");
    }
    if (active == NULL || active->cl_player == NULL) {
	here = NULL;
    } else {
	here = active->cl_player->pl_location;
    }
    cl = Clients;
    while (cl != NULL) {
	pl = cl->cl_player;
	if (! cl->cl_deleted && pl != NULL && cl != active &&
	    (active == NULL || pl->pl_location == here))
	{
	    if (cl->cl_isMachine) {
		runWithString(cl, pl, &PrintMessage[0], "LISTEN_ACTION");
	    } else {
		(void) send(cl->cl_fd, &PrintMessage[0], len, 0);
	    }
	}
	cl = cl->cl_next;
    }
}

void
va_cprintf(Client_t *cl, char *fmt, va_list ap)
{
    int len;

    vsprintf(&PrintMessage[0], fmt, ap);
    len = strlen(&PrintMessage[0]);
    if (len >= LINE_LENGTH) {
	mudAbort("message too long in cprintf");
    }
    if (! cl->cl_isMachine) {
	(void) send(cl->cl_fd, &PrintMessage[0], len, 0);
#if MACHINE_DEBUG
    } else {
	(void) write(1, &PrintMessage[0], len);
#endif
    }
}

void
cprintf(Client_t *cl, char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    va_cprintf(cl, fmt, ap);
    va_end(ap);
}

struct client *
findClient(Thing_t *who)
{
    Client_t *cl;

    cl = Clients;
    while (cl != NULL) {
	if (! cl->cl_deleted && cl->cl_player != NULL &&
	    &cl->cl_player->pl_thing == who)
	{
	    return(cl);
	}
	cl = cl->cl_next;
    }
    return(NULL);
}

Thing_t *
findActivePlayer(char *name, Thing_t *where)
{
    Client_t *cl;

    cl = Clients;
    while (cl != NULL) {
	if (! cl->cl_deleted && cl->cl_player != NULL &&
	    strcmp(cl->cl_player->pl_thing.th_name, name) == 0 &&
	    (where == NULL || cl->cl_player->pl_location == where))
	{
	    return(&cl->cl_player->pl_thing);
	}
	cl = cl->cl_next;
    }
    return(NULL);
}

int
forEachPlayer(char *proc)
{
    Client_t *cl;
    int count;

    count = 0;
    cl = Clients;
    while (cl != NULL) {
	if (! cl->cl_deleted && cl->cl_player != NULL &&
	    cl->cl_state == cs_connected)
	{
	    runExplicitWithPlayer(cl, cl->cl_player, proc);
	    count += 1;
	}
	cl = cl->cl_next;
    }
    return(count);
}

Thing_t *
clientLocation(Thing_t *who)
{
    Client_t *cl;
    Player_t *pl;

    cl = Clients;
    while (cl != NULL) {
	if (! cl->cl_deleted) {
	    pl = cl->cl_player;
	    if (pl != NULL && &pl->pl_thing == who) {
		return(pl->pl_location);
	    }
	}
	cl = cl->cl_next;
    }
    return(NULL);
}

void
setClientLocation(Thing_t *who, Thing_t *location)
{
    Client_t *cl;
    Player_t *pl;

    cl = Clients;
    while (cl != NULL) {
	if (! cl->cl_deleted) {
	    pl = cl->cl_player;
	    if (pl != NULL && &pl->pl_thing == who) {
		pl->pl_location = location;
		return;
	    }
	}
	cl = cl->cl_next;
    }
}

void
setQuit(struct client *cl)
{
    cl->cl_exiting = true;
}

void
initMachine(Player_t *pl)
{
    Client_t *cl;

    cl = (Client_t *) mudAlloc(sizeof(Client_t));
    cl->cl_prev = NULL;
    cl->cl_next = Clients;
    cl->cl_hostName = NULL;
    cl->cl_fd = 2;
    cl->cl_player = pl;
    cl->cl_state = cs_connected;
    cl->cl_exiting = false;
    cl->cl_isMachine = true;
    cl->cl_deleted = false;
    cl->cl_prev = NULL;
    cl->cl_next = Clients;
    Clients = cl;
}

Thing_t *
createMachine(char *name)
{
    Player_t *pl;

    pl = newMachine(name);
    pl->pl_password = strdup("XXX");
    initMachine(pl);
    return(&pl->pl_thing);
}

static void
unlinkClient(Client_t *cl)
{

    if (cl->cl_prev != NULL) {
	cl->cl_prev->cl_next = cl->cl_next;
    } else {
	Clients = cl->cl_next;
    }
    if (cl->cl_next != NULL) {
	cl->cl_next->cl_prev = cl->cl_prev;
    }
    mudFree(cl);
}

Bool_t
destroyMachine(Thing_t *th)
{
    Client_t *cl;
    Action_t *ac, **acp;

    cl = Clients;
    while (cl != NULL &&
	(cl->cl_player == NULL || &cl->cl_player->pl_thing != th))
    {
	cl = cl->cl_next;
    }
    if (cl != NULL) {
	acp = &Actions;
	while ((ac = *acp) != NULL) {
	    if (ac->ac_client == cl) {
		*acp = ac->ac_next;
		mudFree(ac);
	    } else {
		acp = &ac->ac_next;
	    }
	}
	cl->cl_deleted = true;
	return(true);
    } else {
	return(false);
    }
}

void
doAfter(struct client *cl, int delay, char *proc)
{
    Action_t *ac, **acp;

    ac = (Action_t *) mudAlloc(sizeof(Action_t));
    ac->ac_client = cl;
    ac->ac_delay = delay;
    ac->ac_action = proc;
    if (delay < CurrentDelay || Actions == NULL) {
	CurrentDelay = delay;
	LastTime = time(NULL);
	ac->ac_next = Actions;
	Actions = ac;
    } else {
	acp = &Actions;
	while (*acp != NULL && (*acp)->ac_delay <= delay) {
	    acp = &(*acp)->ac_next;
	}
	ac->ac_next = *acp;
	*acp = ac;
    }
}

Bool_t
doForce(Thing_t *who, char *proc)
{
    Client_t *cl;
    Player_t *pl;

    cl = Clients;
    while (cl != NULL) {
	if (! cl->cl_deleted) {
	    pl = cl->cl_player;
	    if (pl != NULL && &pl->pl_thing == who) {
		runExplicitWithPlayer(cl, pl, proc);
		return(true);
	    }
	}
	cl = cl->cl_next;
    }
    return(false);
}

static void
doDelayedActions(void)
{
    int now, delta;
    Action_t *ac;
    Action_t **acp;

    now = time(NULL);
    delta = now - LastTime;
    LastTime = now;
    ac = Actions;
    while (ac != NULL) {
	ac->ac_delay -= delta;
	ac = ac->ac_next;
    }
    acp = &Actions;
    while (true) {
	ac = *acp;
	if (ac == NULL || ac->ac_delay > 0) {
	    break;
	}
	*acp = ac->ac_next;
	runExplicitWithPlayer(ac->ac_client, ac->ac_client->cl_player,
			      ac->ac_action);
	mudFree(ac);
    }
    if (Actions == NULL) {
	CurrentDelay = MAX_DELAY;
    } else {
	CurrentDelay = Actions->ac_delay;
    }
}

static void
startMachines(void)
{
    Client_t *cl;

    cl = Clients;
    while (cl != NULL) {
	if (cl->cl_isMachine) {
	    runWithPlayer(cl, cl->cl_player, "MACHINE_RESTART_ACTION");
	}
	cl = cl->cl_next;
    }
}

static void
cleanupMachines(void)
{
    Client_t **pcl, *cl;

    pcl = &Clients;
    while ((cl = *pcl) != NULL) {
	if (cl->cl_deleted) {
	    if (cl->cl_player != NULL) {
		deleteMachine(cl->cl_player);
	    }
	    unlinkClient(cl);
	} else {
	    pcl = &cl->cl_next;
	}
    }
}

static void
killPlayer(Client_t *cl)
{
    Action_t **acp;
    Action_t *ac;

    killSocket(cl->cl_fd);
    FD_CLR(cl->cl_fd, &AllPlayerBits);
    if (cl->cl_player != NULL) {
	aprintf(cl, "%s@%s has left the game.\n",
		cl->cl_player->pl_thing.th_name, cl->cl_hostName);
    } else {
	aprintf(cl, "???@%s has left the game.\n", cl->cl_hostName);
    }
    acp = &Actions;
    while (true) {
	ac = *acp;
	if (ac == NULL) {
	    break;
	}
	if (ac->ac_client == cl) {
	    *acp = ac->ac_next;
	    mudFree(ac);
	} else {
	    acp = &ac->ac_next;
	}
    }
    unlinkClient(cl);
    PlayerCount -= 1;
}

static void
enterStuff(Client_t *cl, Bool_t isNewPlayer)
{
    Client_t *cl2;
    Bool_t first;

    cprintf(cl, "Welcome to ToyMUD V1.0!\n");
    cl2 = Clients;
    first = true;
    while (cl2 != NULL) {
	if (! cl2->cl_deleted && cl2 != cl && ! cl2->cl_isMachine) {
	    if (first) {
		first = false;
		cprintf(cl, "Currently connected: ");
	    }
	    if (cl2->cl_player == NULL) {
		cprintf(cl, "???@%s ", cl2->cl_hostName);
	    } else {
		cprintf(cl, "%s@%s ", cl2->cl_player->pl_thing.th_name,
			cl2->cl_hostName);
	    }
	}
	cl2 = cl2->cl_next;
    }
    if (first) {
	cprintf(cl, "No-one else is currently connected.\n\n");
    } else {
	cprintf(cl, "\n\n");
    }
    cl->cl_state = cs_connected;
    if (isNewPlayer) {
	runWithPlayer(cl, cl->cl_player, "NEW_PLAYER_ACTION");
    } else {
	runWithPlayer(cl, cl->cl_player, "RESTART_ACTION");
    }
    aprintf(cl, "%s@%s has entered the game.\n",
	    cl->cl_player->pl_thing.th_name, cl->cl_hostName);
}

static void
skipBlanks(void)
{
    char *p;

    p = &UserMessage[ScanPos];
    while (isspace(*p)) {
	++p;
	ScanPos += 1;
    }
}

char *
getWord(void)
{
    char *p, *q;
    char save;

    skipBlanks();
    p = &UserMessage[ScanPos];
    q = p;
    while (*p != '\0' && ! isspace(*p)) {
	++p;
	ScanPos += 1;
    }
    if (*p != '\0') {
	save = *p;
	*p = '\0';
	ScanPos += 1;
	q = strdup(q);
	*p = save;
    } else {
	q = strdup(q);
    }
    return(q);
}

char *
getTail(void)
{
    skipBlanks();
    return(strdup(&UserMessage[ScanPos]));
}

void
setTail(char *st)
{
    if ((int) strlen(st) >= LINE_LENGTH) {
	mudAbort("line too long in setTail");
    }
    strcpy(&UserMessage[0], st);
    ScanPos = 0;
}

static void
processInput(Client_t *cl)
{
    int len, totalLen;
    char *w;
    Player_t *pl;
    Bool_t noProcess;

    len = recv(cl->cl_fd, &UserMessage[0], LINE_LENGTH - 1, 0);
    if (len <= 0) {
	if (cl->cl_state == cs_connected) {
	    runWithPlayer(cl, cl->cl_player, "QUIT_ACTION");
	}
	killPlayer(cl);
	return;
    }
    noProcess = false;
    while (UserMessage[len - 1] == '\n' || UserMessage[len - 1] == '\r') {
	len -= 1;
	if (len == 0) {
	    return;
	}
    }
    if (UserMessage[len - 1] == '-') {
	UserMessage[len - 1] = '\n';
	noProcess = true;
    } else if (UserMessage[len - 1] == '+') {
	len -= 1;
	noProcess = true;
    }
    totalLen = cl->cl_inputLength;
    if (totalLen + len > LINE_LENGTH) {
	len = LINE_LENGTH - totalLen;
    }
    if (len != 0) {
	memcpy(&cl->cl_inputLine[totalLen], &UserMessage[0], len);
	totalLen += len;
    }
    if (noProcess) {
	cl->cl_inputLength = totalLen;
	return;
    }
    memcpy(&UserMessage[0], &cl->cl_inputLine[0], totalLen);
    UserMessage[totalLen] = '\0';
    cl->cl_inputLength = 0;

    ScanPos = 0;
    skipBlanks();
    if (UserMessage[ScanPos] == '\0') {
	if (cl->cl_state != cs_connected) {
	    killPlayer(cl);
	}
	return;
    }

    switch (cl->cl_state) {
    case cs_name:
	w = getWord();
	pl = findPlayer(w);
	if (pl == NULL) {
	    cprintf(cl,
	      "Player '%s' does not exist. Do you wish to create it? ", w);
	    cl->cl_player = newPlayer(strdup(w));
	    cl->cl_state = cs_createQuery;
	} else {
	    cprintf(cl, "Enter player password: ");
	    cl->cl_player = pl;
	    cl->cl_state = cs_password1;
	}
	break;
    case cs_password1:
	w = getWord();
	if (strcmp(w, cl->cl_player->pl_password) == 0) {
	    enterStuff(cl, false);
	} else {
	    cprintf(cl, "Password incorrect, try again: ");
	    cl->cl_state = cs_password2;
	}
	break;
    case cs_password2:
	w = getWord();
	if (strcmp(w, cl->cl_player->pl_password) == 0) {
	    enterStuff(cl, false);
	} else {
	    cprintf(cl, "Password incorrect, try again: ");
	    cl->cl_state = cs_password3;
	}
	break;
    case cs_password3:
	w = getWord();
	if (strcmp(w, cl->cl_player->pl_password) == 0) {
	    enterStuff(cl, false);
	} else {
	    cprintf(cl, "Password incorrect.\n");
	    killPlayer(cl);
	}
	break;
    case cs_connected:
	if (len > 0) {
	    if (UserMessage[ScanPos] == '.') {
		ScanPos += 1;
		skipBlanks();
		runProc(cl, cl->cl_player, &UserMessage[ScanPos]);
	    } else {
		strcpy(&UserMessage[0], &UserMessage[ScanPos]);
		ScanPos = 0;
		runWithString(cl, cl->cl_player, &UserMessage[0],
			      "PARSE_ACTION");
	    }
	}
	if (cl->cl_exiting) {
	    runWithPlayer(cl, cl->cl_player, "QUIT_ACTION");
	    killPlayer(cl);
	}
	break;
    case cs_createQuery:
	w = getWord();
	if (*w == 'y' || *w == 'Y') {
	    cprintf(cl, "Enter new player password: ");
	    cl->cl_state = cs_createPW1;
	} else {
	    deletePlayer(cl->cl_player);
	    killPlayer(cl);
	}
	break;
    case cs_createPW1:
	w = getWord();
	cl->cl_player->pl_password = strdup(w);
	cprintf(cl, "Re-enter password for verification: ");
	cl->cl_state = cs_createPW2;
	break;
    case cs_createPW2:
	w = getWord();
	if (strcmp(cl->cl_player->pl_password, w) == 0) {
	    enterStuff(cl, true);
	} else {
	    cprintf(cl, "Password not verified.\n");
	    deletePlayer(cl->cl_player);
	    killPlayer(cl);
	}
	break;
    }
}

static void
interruptHandler(int sig)
{
    WantQuit = true;
}

int
main(int argc, char *argv[])
{
    struct sockaddr_in myAddr;
    short int hostPort;
    int res;
    Variable_t *globalContext;

#if defined(_WIN32)
    WSADATA WSAData;

    res = WSAStartup(MAKEWORD(1, 0), &WSAData);
    if (res < 0) {
	fprintf(stderr, "WSAStartup failure error %d\n", res);
	exit(1);
    }
#endif
    hostPort = DEFAULT_PORT;
    DbPath = DEFAULT_DB;
    if (argc >= 2) {
	hostPort = atoi(argv[1]);
	if (argc >= 3) {
	    DbPath = argv[2];
	}
    }

    Clients = NULL;
    Actions = NULL;
    PlayerCount = 0;
    globalContext = readDB(DbPath);
    Socket = socket(AF_INET, SOCK_STREAM, 0);
#if defined(_WIN32)
    if (Socket == INVALID_SOCKET) {
	PERROR("socket");
	exit(1);
    }
#else
    if (Socket < 0) {
	perror("socket");
	exit(1);
    }
#endif
    memset(&myAddr, 0, sizeof(struct sockaddr_in));
    myAddr.sin_port = htons(hostPort);
    myAddr.sin_family = AF_INET;
    if (bind(Socket, (struct sockaddr *) &myAddr,
	     sizeof(struct sockaddr_in)) < 0)
    {
	PERROR("bind");
	exit(1);
    }
    if (listen(Socket, 5) < 0) {
	PERROR("listen");
	exit(1);
    }

    FD_ZERO(&AllPlayerBits);
    initInterpreter(globalContext);
    WantQuit = false;
    CurrentDelay = MAX_DELAY;
    LastTime = time(NULL);
    srand(LastTime);
    signal(SIGINT, interruptHandler);
#if defined(unix)
    signal(SIGPIPE, SIG_IGN);
#endif
    startMachines();
    while (! WantQuit || PlayerCount != 0) {
	fd_set readFds, exceptFds;
	struct sockaddr_in from;
	Client_t *cl;
	struct timeval timeval;

	cleanupMachines();
	readFds = AllPlayerBits;
	FD_SET(Socket, &readFds);
	exceptFds = readFds;
	timeval.tv_sec = CurrentDelay;
	timeval.tv_usec = 0;
	res = select(FD_SETSIZE, &readFds, (fd_set *) NULL, &exceptFds,
		     &timeval);
#if defined(_WIN32)
	if (res == SOCKET_ERROR) {
	    if (WSAGetLastError() != WSAEINTR) {
		PERROR("select");
		mudAbort(NULL);
	    }
	    WantQuit = true;
#else
	if (res < 0) {
	    if (errno != EINTR) {
		perror("select");
		mudAbort(NULL);
	    }
	    WantQuit = true;
#endif
	} else if (res == 0) {
	    doDelayedActions();
	} else {
	    if (FD_ISSET(Socket, &exceptFds)) {
		PERROR("select");
		mudAbort(NULL);
	    }
	    if (FD_ISSET(Socket, &readFds)) {
		int clientFd;
#ifdef __SASC
		LONG fromLen;
#else
		int fromLen;
#endif

		fromLen = sizeof(struct sockaddr_in);
		clientFd = accept(Socket, (struct sockaddr *) &from, &fromLen);
#if defined(_WIN32)
		if (clientFd == INVALID_SOCKET) {
		    if (WSAGetLastError() != WSAEINTR) {
			PERROR("accept");
			mudAbort(NULL);
		    }
		} else {
#else
		if (clientFd < 0) {
		    if (errno != EINTR) {
			perror("accept");
			mudAbort(NULL);
		    }
		} else {
#endif
		    struct hostent *hp;

		    cl = (Client_t *) mudAlloc(sizeof(Client_t));
		    cl->cl_fd = clientFd;
		    hp = gethostbyaddr((const char *) &from.sin_addr,
				       sizeof(struct in_addr), AF_INET);
		    if (hp == NULL) {
			sprintf(&PrintMessage[0], "%08lx",
				from.sin_addr.s_addr);
		    } else {
			strcpy(&PrintMessage[0], hp->h_name);
		    }
		    cl->cl_hostName = strdup(&PrintMessage[0]);
		    cl->cl_state = cs_name;
		    cl->cl_player = NULL;
		    cl->cl_next = Clients;
		    if (Clients != NULL) {
			Clients->cl_prev = cl;
		    }
		    cl->cl_prev = NULL;
		    cl->cl_inputLength = 0;
		    cl->cl_exiting = false;
		    cl->cl_isMachine = false;
		    cl->cl_deleted = false;
		    Clients = cl;
		    PlayerCount += 1;
		    FD_SET(clientFd, &AllPlayerBits);
		    cprintf(cl, "Enter player name: ");
		}
	    }
	    cl = Clients;
	    while (cl != NULL) {
		Client_t *nextClient;

		nextClient = cl->cl_next;
		if (FD_ISSET(cl->cl_fd, &exceptFds)) {
		    if (cl->cl_state == cs_createQuery ||
			cl->cl_state == cs_createPW1 ||
			cl->cl_state == cs_createPW2)
		    {
			deletePlayer(cl->cl_player);
			cl->cl_player = NULL;
		    } else {
			runWithPlayer(cl, cl->cl_player, "QUIT_ACTION");
		    }
		    killPlayer(cl);
		} else if (FD_ISSET(cl->cl_fd, &readFds)) {
		    processInput(cl);
		}
		cl = nextClient;
	    }
	    doDelayedActions();
	}
    }

    killSocket(Socket);
    writeDB(DbPath, getGlobals());
#if defined(_WIN32)
    (void) WSACleanup();
#endif
    return(0);
}
